SystemOrganization addCategory: #MetaInterpreter!

TestCase subclass: #TestInterpreter
	instanceVariableNames: 'anIv'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/3/2013 11:50'!
estOnInteractionWithFromDebuggingSupport

	|object collection|

	object := Object new.
	collection := OrderedCollection new.
	
	RunTimeDebuggingSupport onInteractionFor: object with: collection do:  '[:r1 :r2 | 
		r1 trigger halt.
	]'.

	RunTimeDebuggingSupport enableMetaBehaviorForRegistered.
	
	collection add: object.
	
	RunTimeDebuggingSupport removeMetaBehaviorForRegistered.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:13'!
passAsArg: t

! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:39'!
store: t

	anIv := t

! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/22/2013 17:31'!
testCompositeHookOnNew

Interpreter on: MessageReceived for: Point do: (Compiler evaluate: [:reification :reflectogram | 
	reification message selector = #new ifTrue: [
		reification trigger halt.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).	
	]	
] asString).

RunTimeDebuggingSupport evaluate: 'TestClass run'.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/3/2013 11:46'!
testDerivativeHooksWithHalt

	"press proceed for the test to finish"

	|t1 t2|
	
	"register halt on creation"
	Interpreter on: MessageReceived for: TestObject do: [:reifications :reflectogram | 
		reifications message selector = #new ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: TestObject.
	
	t1 := TestObject new. "should halt"
	t2 := TestObject new. "should halt."
	
	Interpreter removeMetaBehaviorOf: TestObject.
	
	"register halt on object in invoke -- this can be done either by registering the argument or the receiver"
	Interpreter on: AsArgument for: t2 do: [:reifications :reflectogram |
			reifications receiver = t1 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t2.
	
	t1 testPassingArgument: t2. "should halt"

	Interpreter removeMetaBehaviorOf: t2.
	
	"register halt on interaction"	
	Interpreter on: MessageReceived for: t1 do: [:reifications :reflectogram | 
		(reifications message arguments includes: t2) ifTrue: [self halt].
	].

	Interpreter on: AsArgument for: t1 do: [:reifications :reflectogram | 
		reifications receiver = t2 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t1.


	t1 testInteraction: t2. "should halt because t1 invokes a method of t2"
	t2 testInteraction: t1. "should halt because t2 invokes a method of t1"
	
	Interpreter disableMetaBehaviorFor: t1.
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:45'!
testHookArg

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:45'!
testHookRead

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivReadCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookReceive

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookSend

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.	
		].
	

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: sendCounter = (2*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:41'!
testHookStore

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter storeCounter |
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.
	storeCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter on: Stored for: t do: [
			storeCounter := storeCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
		self store: t.	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: storeCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookWrite

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivWriteCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:48'!
testHooks

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.
		].
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
		].

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:49'!
testHooksWithHalt

	"press proceed for the test to finish"

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

		
	t := TestObject new.
	
	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		self halt.
		receiveCounter := receiveCounter + 1.
	].
	
	Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
		self halt.
		sendCounter := sendCounter + 1.	
	].
	
	Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
		self halt.
		ivReadCounter := ivReadCounter + 1.
	].

	Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
		self halt.
		ivWriteCounter := ivWriteCounter + 1.
	].

	Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
		self halt.
		argCounter := argCounter + 1.
	].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	
	self assert: receiveCounter = 3.
	self assert: sendCounter = 2.
	self assert: ivReadCounter = 1.
	self assert: ivWriteCounter = 1.
	self assert: argCounter = 1.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:52'!
testHooksWithOverride

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.
		].
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
			reflectogram override: true.
			reflectogram returnValue: 666.
			reflectogram enable.
		].

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 666). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/4/2013 12:07'!
testOnInteractionWithFromDebuggingSupport

	|object collection|

	object := Object new.
	collection := OrderedCollection new.
	
	RunTimeDebuggingSupport onInteractionFor: object with: collection do:  '[:r1 :r2 | 
		r1 trigger halt.
		r2 override: true.
		r2 returnValue: r2 defaultAction.
	]'.

	RunTimeDebuggingSupport enableMetaBehaviorForRegistered.
	
	collection add: object.
	
	RunTimeDebuggingSupport removeMetaBehaviorForRegistered.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:53'!
testOneShot

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 1. "it was one shot"! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:53'!
testOneShotWithPost

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).
		receiveCounter := receiveCounter + 1.
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 2. "it was one shot"! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/10/2013 10:28'!
testSetUnset

	|object|
	
	object := Object new.
	Interpreter enableMetaBehaviorFor: object.
	self assert: (Interpreter isMetaBehaviorEnabledFor: object).
	Interpreter disableMetaBehaviorFor: object.
	self assert: (Interpreter isMetaBehaviorEnabledFor: object) not.! !

TestCase subclass: #TestOptimizedInterpreter
	instanceVariableNames: 'anIv'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
passAsArg: t

! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
store: t

	anIv := t

! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testAugment

	|anObject|
	anObject := Object new.
	OptimizedInterpreter stepAugment: anObject.
	self assert: (anObject instVarAt: anObject class instSize) = nil.
	 (anObject instVarAt: anObject class instSize put: 666).
	self assert: (anObject instVarAt: anObject class instSize) = 666.! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testDerivativeHooksWithHalt

	"press proceed for the test to finish"

	|t1 t2|
	
	"register halt on creation"
	Interpreter on: MessageReceived for: TestObject do: [:reifications :reflectogram | 
		reifications message selector = #new ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: TestObject.
	
	t1 := TestObject new. "should halt"
	t2 := TestObject new. "should halt."
	
	Interpreter removeMetaBehaviorOf: TestObject.
	
	"register halt on object in invoke -- this can be done either by registering the argument or the receiver"
	Interpreter on: AsArgument for: t2 do: [:reifications :reflectogram |
			reifications receiver = t1 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t2.
	
	t1 testPassingArgument: t2. "should halt"

	Interpreter removeMetaBehaviorOf: t2.
	
	"register halt on interaction"	
	Interpreter on: MessageReceived for: t1 do: [:reifications :reflectogram | 
		(reifications message arguments includes: t2) ifTrue: [self halt].
	].

	Interpreter on: AsArgument for: t1 do: [:reifications :reflectogram | 
		reifications receiver = t2 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t1.


	t1 testInteraction: t2. "should halt because t1 invokes a method of t2"
	t2 testInteraction: t1. "should halt because t2 invokes a method of t1"
	
	Interpreter disableMetaBehaviorFor: t1.
! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testHookArg

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: argCounter = (1*100).! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testHookRead

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivReadCounter = (1*100).
! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/13/2013 13:55'!
testHookReceive

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0. 

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		OptimizedInterpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].

		OptimizedInterpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		OptimizedInterpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testHookSend

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.	
		].
	

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: sendCounter = (2*100).
! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testHookStore

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter storeCounter |
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.
	storeCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter on: Stored for: t do: [
			storeCounter := storeCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
		self store: t.	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: storeCounter = (1*100).
! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testHookWrite

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivWriteCounter = (1*100).
! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/13/2013 14:07'!
testHooks

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0. 
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.
		].
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
		].

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testHooksWithHalt

	"press proceed for the test to finish"

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

		
	t := TestObject new.
	
	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		self halt.
		receiveCounter := receiveCounter + 1.
	].
	
	Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
		self halt.
		sendCounter := sendCounter + 1.	
	].
	
	Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
		self halt.
		ivReadCounter := ivReadCounter + 1.
	].

	Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
		self halt.
		ivWriteCounter := ivWriteCounter + 1.
	].

	Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
		self halt.
		argCounter := argCounter + 1.
	].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	
	self assert: receiveCounter = 3.
	self assert: sendCounter = 2.
	self assert: ivReadCounter = 1.
	self assert: ivWriteCounter = 1.
	self assert: argCounter = 1.! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testHooksWithOverride

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.
		].
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
			reflectogram override: true.
			reflectogram returnValue: 666.
			reflectogram enable.
		].

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 666). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/13/2013 13:33'!
testMetaFetch

	|anObject|
	anObject := Object new.
	self assert: (OptimizedInterpreter metaObjectOf: anObject) class = LowLevelImplicitMetaObject.
	! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testOnInteractionWithFromDebuggingSupport

	|object collection|

	object := Object new.
	collection := OrderedCollection new.
	
	RunTimeDebuggingSupport onInteractionFor: object with: collection do:  '[:r1 :r2 | 
		r1 trigger halt.
		r2 override: true.
		r2 returnValue: r2 defaultAction.
	]'.

	RunTimeDebuggingSupport enableMetaBehaviorForRegistered.
	
	collection add: object.
	
	RunTimeDebuggingSupport removeMetaBehaviorForRegistered.! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testOneShot

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 1. "it was one shot"! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testOneShotWithPost

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).
		receiveCounter := receiveCounter + 1.
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 2. "it was one shot"! !

!TestOptimizedInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 16:08'!
testSetUnset

	|object|
	
	object := Object new.
	Interpreter enableMetaBehaviorFor: object.
	self assert: (Interpreter isMetaBehaviorEnabledFor: object).
	Interpreter disableMetaBehaviorFor: object.
	self assert: (Interpreter isMetaBehaviorEnabledFor: object) not.! !

Object subclass: #BenchmarkObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!BenchmarkObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
test
	| a b c |
	a := 1.
	b := 2.
	c := a + b.
	^ c
! !

Object subclass: #LazyReifications
	instanceVariableNames: 'interpreterContext'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LazyReifications class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 10:54'!
with: anInterpreterContext

	^ self new interpreterContext: anInterpreterContext! !

!LazyReifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:10'!
interpreterContext

	^ interpreterContext! !

!LazyReifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:10'!
interpreterContext: aContext

	interpreterContext := aContext! !

LazyReifications subclass: #Reifications
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:11'!
context

	^ interpreterContext sender ! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
index

	^ interpreterContext tempNamed: #anIndex! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
message

	^ self interpreterContext tempNamed: #aMessage
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
method

	^ self context method! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
receiver

	^ interpreterContext tempNamed: #rcvr
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
sender

	^ self context receiver! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
sourceRange

	^ self context debuggerMap 
						rangeForPC: self context pc
						contextIsActiveContext: false 
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
trigger

	^ interpreterContext tempNamed: #anObject! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:14'!
value

	^ interpreterContext tempNamed: #aValue! !

Object subclass: #LowLevelImplicitMetaObject
	instanceVariableNames: 'receiveCallback sendCallback readCallback writeCallback argCallback storeCallback'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LowLevelImplicitMetaObject class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/7/2013 17:57'!
new

	^ self basicNew initialize! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback
	^ argCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback: anObject
	argCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/2/2013 15:16'!
initialize 

	super initialize.

	"self receiveCallback: [:reifications :override | ].

	self sendCallback: [:reifications :override | ].

	self readCallback: [:reifications :override | ].

	self writeCallback: [:reifications :override | ].
	
	self	 argCallback: [:reifications :override | ].
		"! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback
	^ readCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback: anObject
	readCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback
	^ receiveCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback: anObject
	receiveCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback
	^ sendCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback: anObject
	sendCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/17/2013 11:29'!
storeCallback
	^ storeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/17/2013 11:29'!
storeCallback: anObject
	storeCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback
	^ writeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback: anObject
	writeCallback := anObject! !

Object subclass: #MetaBenchmark
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/15/2013 14:47'!
benchmark: aBlock times: anInteger
	"synced with Bifrost's bench"
	| runs average deviation |
	aBlock value; value. "this does nothing in our case possible bifrost uses it
	to install methodWrappers the first time this is run ?"
	runs := [ 
		(1 to: anInteger) 
			collect: [ :each | aBlock timeToRun ] ]
		valueUnpreemptively.
	average := runs average asFloat.
	deviation := ((runs 
		detectSum: [ :each | (each - average) * (each - average) ]) 
			/ runs size) sqrt.
	^ Array with: average with: deviation! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
message
	"self new message" 
	
	|  object  |
	object := BenchmarkObject new.
	^ self benchmark: [ 1 to: 1000000 do: [ :each | object test ] ] times: 100.! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/5/2013 11:57'!
messageMetaJumpWithHaltOnly
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	HaltInterpreter initialize.
	 
	HaltInterpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	HaltInterpreter removeMetaBehaviorOf: object.
	
	Interpreter initialize.
	
	^ result! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 17:10'!
messageMetaSend
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	Interpreter on: MessageReceived for: object do: [].
	 
	Interpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	Interpreter removeMetaBehaviorOf: object.
	
	^ result! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 17:51'!
messageMetaSendWithReifications
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	Interpreter on: MessageReceived for: object do: [:reifications :reflectogram | ].
	 
	Interpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	Interpreter removeMetaBehaviorOf: object.
	
	^ result! !

Object subclass: #MetaHandler
	instanceVariableNames: 'baseObject implicitMetaObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MetaHandler methodsFor: 'accessing' stamp: 'NickPapoulias 10/7/2013 16:45'!
baseObject
	^ baseObject! !

!MetaHandler methodsFor: 'accessing' stamp: 'NickPapoulias 10/7/2013 16:45'!
baseObject: anObject
	baseObject := anObject! !

!MetaHandler methodsFor: 'accessing' stamp: 'NickPapoulias 10/7/2013 16:45'!
implicitMetaObject
	^ implicitMetaObject! !

!MetaHandler methodsFor: 'accessing' stamp: 'NickPapoulias 10/7/2013 16:45'!
implicitMetaObject: anObject
	implicitMetaObject := anObject! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit
	^ self primitiveGetExperimentalBit ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit: aBoolean
	self primitiveSetExperimentalBitTo: aBoolean! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/8/2013 15:34'!
identityHash
	"Answer a SmallInteger whose value is related to the receiver's identity.
	This method must not be overridden, except by SmallInteger.
	
	Do not override."

	^ (Interpreter object: self perform: #basicIdentityHash withArguments: {}) bitShift: 18! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
isExperimentalBitOn
	^ self experimentalBit! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaArgIn: aMessage for: anObject

	^ Interpreter object: anObject metaArgIn: aMessage withReceiver: self! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReadAt: anIndex

	^ Interpreter object: self metaReadAt: anIndex! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReceived: aMessage

	^ Interpreter object: self metaReceived: aMessage ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:39'!
metaRef

	self halt! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaSend: aMessage from: anObject

	^ Interpreter object: anObject metaSend: aMessage withReceiver: self.! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaWriteAt: anIndex Put: anotherObject

	^ Interpreter object: self metaWriteAt: anIndex Put: anotherObject! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveGetExperimentalBit

	<primitive: 150>
	self primitiveFailed! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveSetExperimentalBitTo: aBoolean

	<primitive: 151>
	self primitiveFailed! !

Object subclass: #Operation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #AsArgument
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!AsArgument class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject argCallback: aCallback ! !

Operation subclass: #AsReturnValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #CompositeOperation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

CompositeOperation subclass: #ObjectCreation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

CompositeOperation subclass: #ObjectInteraction
	instanceVariableNames: 'interactionObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!ObjectInteraction class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/22/2013 18:31'!
with: anotherObject

	^ self new with: anotherObject ! !

!ObjectInteraction methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/3/2013 11:51'!
registerInteractionCallback: aCallback for: anObject

	Interpreter on: AsArgument for: anObject do: [:reifications :reflectogram | 
		reifications receiver = interactionObject ifTrue: [aCallback value: reifications value: reflectogram]	
	].
	
	Interpreter on: MessageReceived for: anObject  do: [:reifications :reflectogram | 
		(reifications message arguments includes: interactionObject) ifTrue: [aCallback value: reifications value: reflectogram]	
	].
! !

!ObjectInteraction methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/2/2013 14:03'!
with: anotherObject

	interactionObject := anotherObject! !

Operation subclass: #IvRead
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!IvRead class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject readCallback: aCallback ! !

Operation subclass: #IvWrite
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!IvWrite class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:23'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject writeCallback: aCallback ! !

Operation subclass: #MessageReceived
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MessageReceived class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/9/2013 17:02'!
registerCallback: aCallback to: anImplicitMetaObject



	anImplicitMetaObject receiveCallback: aCallback ! !

Operation subclass: #MessageSent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MessageSent class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:24'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject sendCallback: aCallback ! !

!Operation class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:15'!
registerCallback: aCallback to: anImplicitMetaObject

	self subclassResponsibility ! !

Operation subclass: #Referenced
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #Stored
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Stored class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject storeCallback: aCallback ! !

Object subclass: #Reflectogram
	instanceVariableNames: 'trigger returnValue defaultAction override'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reflectogram class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:18'!
for: anObject withDefaultAction: aBlock

	^ self basicNew initializeWithTrigger: anObject andAction: aBlock! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:56'!
cascade: aBlock

	| ret |
	Interpreter enableMetaBehaviorFor: trigger.
	ret := aBlock value.
	Interpreter disableMetaBehaviorFor: trigger.
	^ ret.! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:07'!
defaultAction

	^ defaultAction value! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
disable 

	Interpreter disableMetaBehaviorFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:56'!
enable

	Interpreter enableMetaBehaviorFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:45'!
initializeWithTrigger: anObject andAction: aBlock

	trigger := anObject.
	defaultAction := aBlock.
	override := false.
	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
isEnabled 

	^ Interpreter isMetaBehaviorEnabledFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:44'!
override

	^ override ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:43'!
override: aBoolean

	override := aBoolean ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/22/2013 18:12'!
remove

	Interpreter removeMetaBehaviorOf: trigger.! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
returnValue

	^ returnValue ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
returnValue: aValue

	returnValue := aValue! !

Object subclass: #TestObject
	instanceVariableNames: 'x'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 16:04'!
testInteraction: anObject 

	anObject class.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 15:58'!
testPassingArgument: anObject ! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/9/2013 19:38'!
testSenderHook

	Transcript show: 'this message should be caught'; cr.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:01'!
x

	^ x! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/5/2013 19:02'!
x: anObject 

	x := anObject ! !

Interpreter subclass: #OptimizedInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!OptimizedInterpreter class methodsFor: 'initialize' stamp: 'NickPapoulias 10/13/2013 13:58'!
initialize 

	super initialize.
	Smalltalk globals at: #Interpreter put: self.! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/13/2013 13:23'!
metaObjectOf: anObject

	|lastIndex|
	
	[lastIndex := (anObject instVarAt: anObject class instSize)] on: Error do: [
		self stepAugment: anObject.
		^ (anObject instVarAt: anObject class instSize put: LowLevelImplicitMetaObject new).			
	].

	lastIndex class = LowLevelImplicitMetaObject ifTrue: [^ lastIndex ] ifFalse: [
		self stepAugment: anObject.
		^ (anObject instVarAt: anObject class instSize put: LowLevelImplicitMetaObject new).			
	].
	

	! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/13/2013 15:08'!
object: anObject metaArgIn: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (anObject instVarAt: anObject class instSize) argCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:  [Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments.				
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/13/2013 15:08'!
object: anObject metaReadAt: anIndex

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback :=(anObject instVarAt: anObject class instSize) readCallback.	 
	
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anObject at: anIndex.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: anObject at: anIndex. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: anObject at: anIndex]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject at: anIndex.				
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/13/2013 15:08'!
object: anObject metaReceived: aMessage

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (anObject instVarAt: anObject class instSize) receiveCallback.	
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:  [ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments.				
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/13/2013 15:08'!
object: anObject metaSend: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (anObject instVarAt: anObject class instSize) sendCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.	
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/13/2013 15:08'!
object: anObject metaStoreAt: anIndex for: anotherObject

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (anObject instVarAt: anObject class instSize) storeCallback.	
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^   Interpreter object: anotherObject at: anIndex put: anObject
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anotherObject at: anIndex put: anObject
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [ Interpreter object: anotherObject at: anIndex put: anObject]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anotherObject at: anIndex put: anObject.
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/13/2013 15:08'!
object: anObject metaWriteAt: anIndex Put: aValue

	| reflectogram callback | 
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (anObject instVarAt: anObject class instSize) writeCallback.	
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^   Interpreter object: anObject at: anIndex put: aValue
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject at: anIndex put: aValue
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [ Interpreter object: anObject at: anIndex put: aValue]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anObject at: anIndex put: aValue
	]. ! !

!OptimizedInterpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 10/13/2013 13:43'!
on: anOperationOrEvent for: anObject do: aCallback

	| implicitMetaObject |
	
	anOperationOrEvent class = ObjectInteraction ifTrue: [
		anOperationOrEvent registerInteractionCallback: aCallback  for: anObject 
	] ifFalse: [
	
		implicitMetaObject := (self metaObjectOf: anObject).
 
		anOperationOrEvent registerCallback: aCallback to: implicitMetaObject 
	]! !

!OptimizedInterpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 10/13/2013 13:44'!
onArgumentForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := (self metaObjectOf: anObject).
	implicitMetaObject argCallback: aCallback.
! !

!OptimizedInterpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 10/13/2013 13:44'!
onIvReadForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject :=  self metaObjectOf: anObject.

	implicitMetaObject readCallback: aCallback.
! !

!OptimizedInterpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 10/13/2013 13:45'!
onIvWriteForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := self metaObjectOf: anObject.
	implicitMetaObject writeCallback: aCallback.! !

!OptimizedInterpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 10/13/2013 13:46'!
onMessageReceivedForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := self metaObjectOf: anObject.
	implicitMetaObject receiveCallback: aCallback.
! !

!OptimizedInterpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 10/13/2013 13:46'!
onMessageSendFromObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := self metaObjectOf: anObject.
	implicitMetaObject sendCallback: aCallback.
! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/9/2013 16:04'!
shadowSubclassFor: anObject

	"todo: compile class and size to return the superclass"

	^ anObject class
		subclass: ('Shadow' , anObject hash asString) asSymbol 
		instanceVariableNames: 'implicitMetaObject'
		classVariableNames: ''
		poolDictionaries: ''
		category: 'Unclassified'.	! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/9/2013 16:16'!
stepAugment: anObject

	"called a single time for each intercessed object -- shallow copies the object to an instance of a shadowClass
	that has one additional slot for storing the implicit meta-object"

	| class newObject index |
	
	class := anObject class.
	class isVariable
		ifTrue: 
			[index := anObject basicSize.
			newObject := (self variableShadowSubclassFor: anObject) basicNew: index. 
			[index > 0]
				whileTrue: 
					[newObject basicAt: index put: (anObject basicAt: index).
					index := index - 1].
			]
		ifFalse: [newObject := (self shadowSubclassFor: anObject) basicNew]. 
	index := class instSize.
	[index > 0]
		whileTrue: 
			[newObject instVarAt: index put: (anObject instVarAt: index).
			index := index - 1].
		
	^ anObject becomeForward: newObject ! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/9/2013 16:01'!
variableShadowSubclassFor: anObject

	
	^ anObject class
		variableSubclass: ('Shadow' , anObject hash asString) asSymbol 
		instanceVariableNames: 'implicitMetaObject'
		classVariableNames: ''
		poolDictionaries: ''
		category: 'Unclassified'.		

	! !
OptimizedInterpreter initialize!
